@@ -1,12 +1,18 @@
 from io import BytesIO
-from typing import Any, ClassVar, Dict, Optional
+from typing import TYPE_CHECKING, Any, Callable, ClassVar, Dict, Optional
 from uuid import UUID
 
+from google.protobuf.internal.containers import MessageMap
+
 from .node import Node
 from .proto import AuxData_pb2
 from .serialization import Serialization
 from .util import DictLike
 
+if TYPE_CHECKING:  # pragma: no cover
+    # Ignore flake8 "imported but unused" errors.
+    from .ir import IR  # noqa: F401
+
 
 class _LazyDataContainer:
     """
@@ -15,13 +21,17 @@ class _LazyDataContainer:
     deserializing) the data.
     """
 
-    def __init__(self, raw_data, type_name, get_by_uuid):
-        self.raw_data = raw_data
+    def __init__(
+        self,
+        raw_data: bytes,
+        type_name: str,
+        get_by_uuid: Callable[[UUID], Optional[Node]],
+    ):
+        self.raw_data: Optional[bytes] = raw_data
         self.type_name = type_name
         self.get_by_uuid = get_by_uuid
 
-    def get_data(self):
-        # type: (Any) -> Any
+    def get_data(self) -> object:
         """
         Get any pending still-serialized data, or return the passed data
         instead (the default).
@@ -33,8 +43,7 @@ class _LazyDataContainer:
         self.raw_data = None
         return rv
 
-    def get_raw_data(self):
-        # type: () -> bytes
+    def get_raw_data(self) -> bytes:
         """
         """
         assert self.raw_data is not None
@@ -59,14 +68,18 @@ class AuxData:
         Used to determine the proper codec for serializing this AuxData.
     """
 
-    serializer = Serialization()  # type: ClassVar[Serialization]
+    serializer: ClassVar[Serialization] = Serialization()
     """This is a :class:`gtirb.Serialization` instance, used to
     encode and decode ``data`` fields of all ``AuxData``. See
     :mod:`gtirb.serialization` for details.
     """
 
-    def __init__(self, data, type_name, lazy_container=None):
-        # type: (Any, str, Optional[_LazyDataContainer]) -> None
+    def __init__(
+        self,
+        data: object,
+        type_name: str,
+        lazy_container: Optional[_LazyDataContainer] = None,
+    ):
         """
         :param data: The value stored in this AuxData.
         :param type_name: A string describing the type of ``data``.
@@ -75,24 +88,29 @@ class AuxData:
             auxdata table backing this object, or None.
         """
         self._lazy_container = lazy_container
-        self._data = data  # type: Any
-        self.type_name = type_name  # type: str
+        # _data has type Any to avoid disrupting clients want to type check
+        # their use of gtirb. If _data had type object, they would have to
+        # verify the element types of potentially large containers, or else
+        # just subvert the type system by casting anyway.
+        self._data: Any = data  # type: ignore[misc]
+        self.type_name = type_name
 
     @property
-    def data(self):
+    def data(self) -> Any:  # type: ignore[misc]
         if self._lazy_container is not None:
             self._data = self._lazy_container.get_data()
             self._lazy_container = None
         return self._data
 
     @data.setter
-    def data(self, value):
+    def data(self, value: object) -> None:
         self._data = value
         self._lazy_container = None
 
     @classmethod
-    def _from_protobuf(cls, aux_data, ir):
-        # type: (AuxData_pb2.AuxData, Optional["IR"]) -> AuxData
+    def _from_protobuf(
+        cls, aux_data: AuxData_pb2.AuxData, ir: Optional["IR"],
+    ) -> "AuxData":
         """Deserialize AuxData from Protobuf. Lazy, will not perform
         deserialization until .data is accessed.
 
@@ -100,6 +118,7 @@ class AuxData:
         """
 
         # Defer deserialization until someone accesses .data
+        assert ir
         lazy_container = _LazyDataContainer(
             aux_data.data, aux_data.type_name, ir.get_by_uuid
         )
@@ -109,8 +128,7 @@ class AuxData:
             lazy_container=lazy_container,
         )
 
-    def _to_protobuf(self):
-        # type: () -> AuxData_pb2.AuxData
+    def _to_protobuf(self) -> AuxData_pb2.AuxData:
         """Get a Protobuf representation of the AuxData."""
 
         proto_auxdata = AuxData_pb2.AuxData()
@@ -127,8 +145,7 @@ class AuxData:
             proto_auxdata.data = data_stream.getvalue()
         return proto_auxdata
 
-    def __repr__(self):
-        # type: () -> str
+    def __repr__(self) -> str:
         return (
             "AuxData("
             "type_name={type_name!r}, "
@@ -148,10 +165,9 @@ class AuxDataContainer(Node):
 
     def __init__(
         self,
-        aux_data={},  # type: DictLike[str, AuxData]
-        uuid=None,  # type: Optional[UUID]
+        aux_data: DictLike[str, AuxData] = {},
+        uuid: Optional[UUID] = None,
     ):
-        # type: (...) -> None
         """
         :param aux_data: The initial auxiliary data to be associated
             with the object, as a mapping from names to
@@ -161,11 +177,14 @@ class AuxDataContainer(Node):
             Defaults to None.
         """
         super().__init__(uuid)
-        self.aux_data = dict(aux_data)  # type: Dict[str, AuxData]
+        self.aux_data: Dict[str, AuxData] = dict(aux_data)
 
     @classmethod
-    def _read_protobuf_aux_data(cls, proto_container, ir):
-        # type: (Any,Optional["IR"]) -> Dict[str, AuxData]
+    def _read_protobuf_aux_data(
+        cls,
+        proto_container: "MessageMap[str, AuxData_pb2.AuxData]",
+        ir: Optional["IR"],
+    ) -> Dict[str, AuxData]:
         """
         Instead of the overrided _decode_protobuf, this method requires the
         Protobuf message to read from. AuxDataContainers need to call this
@@ -176,11 +195,12 @@ class AuxDataContainer(Node):
         """
         return {
             key: AuxData._from_protobuf(val, ir)
-            for key, val in proto_container.aux_data.items()
+            for key, val in proto_container.items()
         }
 
-    def _write_protobuf_aux_data(self, proto_container):
-        # type: (Any) -> None
+    def _write_protobuf_aux_data(
+        self, proto_container: "MessageMap[str, AuxData_pb2.AuxData]"
+    ) -> None:
         """
         Instead of the overrided _to_protobuf, this method requires the
         Protobuf message to write into. AuxDataContainers need to call this
@@ -190,10 +210,9 @@ class AuxDataContainer(Node):
             ``aux_data``.
         """
         for k, v in self.aux_data.items():
-            proto_container.aux_data[k].CopyFrom(v._to_protobuf())
+            proto_container[k].CopyFrom(v._to_protobuf())
 
-    def deep_eq(self, other):
-        # type: (Any) -> bool
+    def deep_eq(self, other: object) -> bool:
         """This overrides :func:`gtirb.Node.deep_eq` to check for
         AuxData equality.
 
