@@ -3,11 +3,23 @@ import typing
 from enum import Enum
 from uuid import UUID
 
+from intervaltree import IntervalTree
+
 from .block import ByteBlock, CodeBlock, DataBlock
 from .byteinterval import ByteInterval, SymbolicExpressionElement
-from .node import Node
+from .node import Node, _NodeMessage
 from .proto import Section_pb2
-from .util import SetWrapper, nodes_at, nodes_on, symbolic_expressions_at
+from .util import (
+    SetWrapper,
+    _address_interval,
+    _nodes_at_interval_tree,
+    _nodes_on_interval_tree,
+)
+
+if typing.TYPE_CHECKING:  # pragma: no cover
+    # Ignore flake8 "imported but unused" errors.
+    from .ir import IR  # noqa: F401
+    from .module import Module  # noqa: F401
 
 
 class Section(Node):
@@ -45,22 +57,26 @@ class Section(Node):
         ThreadLocal = Section_pb2.SectionFlag.Value("ThreadLocal")
         """This section is created in memory once per thread."""
 
-    class _ByteIntervalSet(SetWrapper):
-        def __init__(self, node, *args):
-            self._node = node  # type: Section
+    class _ByteIntervalSet(SetWrapper[ByteInterval]):
+        def __init__(
+            self, node: "Section", *args: typing.Iterable[ByteInterval]
+        ):
+            self._node = node
             super().__init__(*args)
 
-        def add(self, v):
+        def add(self, v: ByteInterval) -> None:
             if v._section is not None:
                 v._section.byte_intervals.discard(v)
+            self._node._index_add(v)
             v._section = self._node
             if self._node.ir is not None:
                 v._add_to_uuid_cache(self._node.ir._local_uuid_cache)
             return super().add(v)
 
-        def discard(self, v):
+        def discard(self, v: ByteInterval) -> None:
             if v not in self:
                 return
+            self._node._index_discard(v)
             v._section = None
             if self._node.ir is not None:
                 v._remove_from_uuid_cache(self._node.ir._local_uuid_cache)
@@ -69,11 +85,11 @@ class Section(Node):
     def __init__(
         self,
         *,
-        name="",  # type: str
-        byte_intervals=(),  # type: typing.Iterable[ByteInterval]
-        flags=set(),  # type: typing.Iterable[Section.Flag]
-        uuid=None,  # type: typing.Optional[UUID]
-        module=None  # type: typing.Optional["Module"]
+        name: str = "",
+        byte_intervals: typing.Iterable[ByteInterval] = (),
+        flags: typing.Iterable["Section.Flag"] = set(),
+        uuid: typing.Optional[UUID] = None,
+        module: typing.Optional["Module"] = None,
     ):
         """
         :param name: The name of this section.
@@ -86,19 +102,34 @@ class Section(Node):
         """
 
         super().__init__(uuid)
-        self._module = None  # type: "Module"
-        self.name = name  # type: str
-        self.byte_intervals = Section._ByteIntervalSet(
-            self, byte_intervals
-        )  # type: typing.Set[ByteInterval]
-        self.flags = set(flags)  # type: typing.Set[Section.Flag]
+        self._interval_index: "IntervalTree[int,ByteInterval]" = IntervalTree()
+        self._module: typing.Optional["Module"] = None
+        self.name = name
+        self.byte_intervals = Section._ByteIntervalSet(self, byte_intervals)
+        self.flags = set(flags)
 
         # Use the property setter to ensure correct invariants.
         self.module = module
 
+    def _index_add(self, byte_interval: ByteInterval) -> None:
+        address_interval = _address_interval(byte_interval)
+        if address_interval:
+            self._interval_index.add(address_interval)
+
+    def _index_discard(self, byte_interval: ByteInterval) -> None:
+        address_interval = _address_interval(byte_interval)
+        if address_interval:
+            self._interval_index.discard(address_interval)
+
     @classmethod
-    def _decode_protobuf(cls, proto_section, uuid, ir):
-        # type: (Section_pb2.Section, UUID, typing.Optional["IR"]) -> Section
+    def _decode_protobuf(
+        cls,
+        proto_section: _NodeMessage,
+        uuid: UUID,
+        ir: typing.Optional["IR"],
+    ) -> "Section":
+        assert ir
+        assert isinstance(proto_section, Section_pb2.Section)
         s = cls(
             name=proto_section.name,
             flags=(Section.Flag(f) for f in proto_section.section_flags),
@@ -111,8 +142,7 @@ class Section(Node):
         )
         return s
 
-    def _to_protobuf(self):
-        # type: () -> Section_pb2.Section
+    def _to_protobuf(self) -> Section_pb2.Section:
         """Get a Protobuf representation of ``self``."""
 
         proto_section = Section_pb2.Section()
@@ -124,8 +154,7 @@ class Section(Node):
         proto_section.section_flags.extend(f.value for f in self.flags)
         return proto_section
 
-    def deep_eq(self, other):
-        # type: (typing.Any) -> bool
+    def deep_eq(self, other: object) -> bool:
         # Do not move __eq__. See docstring for Node.deep_eq for more info.
         if not isinstance(other, Section):
             return False
@@ -143,8 +172,7 @@ class Section(Node):
             and self.flags == other.flags
         )
 
-    def __repr__(self):
-        # type: () -> str
+    def __repr__(self) -> str:
         return (
             "Section("
             "uuid={uuid!r}, "
@@ -160,23 +188,20 @@ class Section(Node):
         )
 
     @property
-    def module(self):
-        # type: () -> "Module"
+    def module(self) -> typing.Optional["Module"]:
         """The :class:`Module` this section belongs to."""
 
         return self._module
 
     @module.setter
-    def module(self, value):
-        # type: ("Module") -> None
+    def module(self, value: typing.Optional["Module"]) -> None:
         if self._module is not None:
             self._module.sections.discard(self)
         if value is not None:
             value.sections.add(self)
 
     @property
-    def byte_blocks(self):
-        # type: () -> typing.Iterator[ByteBlock]
+    def byte_blocks(self) -> typing.Iterator[ByteBlock]:
         """The :class:`ByteBlock`\\s in this section."""
 
         return itertools.chain.from_iterable(
@@ -184,22 +209,19 @@ class Section(Node):
         )
 
     @property
-    def code_blocks(self):
-        # type: () -> typing.Iterator[CodeBlock]
+    def code_blocks(self) -> typing.Iterator[CodeBlock]:
         """The :class:`CodeBlock`\\s in this section."""
 
         return (b for b in self.byte_blocks if isinstance(b, CodeBlock))
 
     @property
-    def data_blocks(self):
-        # type: () -> typing.Iterator[DataBlock]
+    def data_blocks(self) -> typing.Iterator[DataBlock]:
         """The :class:`DataBlock`\\s in this section."""
 
         return (b for b in self.byte_blocks if isinstance(b, DataBlock))
 
     @property
-    def address(self):
-        # type: () -> typing.Optional[int]
+    def address(self) -> typing.Optional[int]:
         """Get the address of this section, if known.
 
         The address is calculated from the :class:`ByteInterval` objects in
@@ -211,17 +233,13 @@ class Section(Node):
         size, so it will be ``None`` in that case.
         """
 
-        lowest = None
-        for bi in self.byte_intervals:
-            if bi.address is None:
-                return None
-            if lowest is None or bi.address < lowest:
-                lowest = bi.address
-        return lowest
+        if 0 < len(self._interval_index) == len(self.byte_intervals):
+            return self._interval_index.begin()
+
+        return None
 
     @property
-    def size(self):
-        # type: () -> typing.Optional[int]
+    def size(self) -> typing.Optional[int]:
         """Get the size of this section, if known.
 
         The address is calculated from the :class:`ByteInterval` objects in
@@ -233,103 +251,120 @@ class Section(Node):
         it has no address or size, so it will be ``None`` in that case.
         """
 
-        lowest = None
-        highest = None
-        for bi in self.byte_intervals:
-            if bi.address is None:
-                return None
-            if lowest is None or bi.address < lowest:
-                lowest = bi.address
-            if highest is None or bi.address + bi.size > highest:
-                highest = bi.address + bi.size
-        if lowest is None or highest is None:
-            return None
-        return highest - lowest
+        if 0 < len(self._interval_index) == len(self.byte_intervals):
+            return self._interval_index.span() - 1
 
-    def byte_intervals_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[ByteInterval]
+        return None
+
+    def byte_intervals_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[ByteInterval]:
         """Finds all the byte intervals that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.byte_intervals, addrs)
+        return _nodes_on_interval_tree(self._interval_index, addrs)
 
-    def byte_intervals_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[ByteInterval]
+    def byte_intervals_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[ByteInterval]:
         """Finds all the byte intervals that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.byte_intervals, addrs)
+        return _nodes_at_interval_tree(self._interval_index, addrs)
 
-    def byte_blocks_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
+    def byte_blocks_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[ByteBlock]:
         """Finds all the byte blocks that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.byte_blocks, addrs)
+        for interval in self.byte_intervals_on(addrs):
+            yield from interval.byte_blocks_on(addrs)
 
-    def byte_blocks_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
+    def byte_blocks_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[ByteBlock]:
         """Finds all the byte blocks that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.byte_blocks, addrs)
+        for interval in self.byte_intervals_on(addrs):
+            yield from interval.byte_blocks_at(addrs)
 
-    def code_blocks_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
+    def code_blocks_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[CodeBlock]:
         """Finds all the code blocks that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.code_blocks, addrs)
+        return (
+            block
+            for block in self.byte_blocks_on(addrs)
+            if isinstance(block, CodeBlock)
+        )
 
-    def code_blocks_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
+    def code_blocks_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[CodeBlock]:
         """Finds all the code blocks that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.code_blocks, addrs)
+        return (
+            block
+            for block in self.byte_blocks_at(addrs)
+            if isinstance(block, CodeBlock)
+        )
 
-    def data_blocks_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
+    def data_blocks_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[DataBlock]:
         """Finds all the data blocks that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.data_blocks, addrs)
+        return (
+            block
+            for block in self.byte_blocks_on(addrs)
+            if isinstance(block, DataBlock)
+        )
 
-    def data_blocks_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
+    def data_blocks_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[DataBlock]:
         """Finds all the data blocks that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.data_blocks, addrs)
+        return (
+            block
+            for block in self.byte_blocks_at(addrs)
+            if isinstance(block, DataBlock)
+        )
 
     def symbolic_expressions_at(
-        self, addrs  # type: typing.Union[int, range]
-    ):
-        # type: (...) -> typing.Iterable[SymbolicExpressionElement]
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[SymbolicExpressionElement]:
         """Finds all the symbolic expressions that begin at an address or
         range of addresses.
 
@@ -338,18 +373,17 @@ class Section(Node):
             symbolic expression in the range.
         """
 
-        return symbolic_expressions_at(self.byte_intervals, addrs)
+        for interval in self.byte_intervals_on(addrs):
+            yield from interval.symbolic_expressions_at(addrs)
 
-    def _add_to_uuid_cache(self, cache):
-        # type: (typing.Dict[UUID, Node]) -> None
+    def _add_to_uuid_cache(self, cache: typing.Dict[UUID, Node]) -> None:
         """Update the UUID cache when this node is added."""
 
         cache[self.uuid] = self
         for bi in self.byte_intervals:
             bi._add_to_uuid_cache(cache)
 
-    def _remove_from_uuid_cache(self, cache):
-        # type: (typing.Dict[UUID, Node]) -> None
+    def _remove_from_uuid_cache(self, cache: typing.Dict[UUID, Node]) -> None:
         """Update the UUID cache when this node is removed."""
 
         del cache[self.uuid]
@@ -357,8 +391,7 @@ class Section(Node):
             bi._remove_from_uuid_cache(cache)
 
     @property
-    def ir(self):
-        # type: () -> "IR"
+    def ir(self) -> typing.Optional["IR"]:
         """Get the IR this node ultimately belongs to."""
         if self.module is None:
             return None
