@@ -5,6 +5,9 @@ from uuid import UUID
 from .node import Node
 from .proto import SymbolicExpression_pb2
 from .symbol import Symbol
+from .util import DeserializationError
+
+AttributesCtorType = typing.Iterable["SymbolicExpression.Attribute"]
 
 
 class SymbolicExpression:
@@ -54,22 +57,22 @@ class SymbolicExpression:
         )
 
     def __init__(
-        self, attributes=set(),
+        self, attributes: AttributesCtorType = set(),
     ):
-        # type: ("AttributesCtorType") -> None
         self.attributes = set(attributes)
 
     @property
-    def symbols(self):
-        # type: () -> typing.Iterable[Symbol]
+    def symbols(self) -> typing.Iterable[Symbol]:
         """Get all the symbols involved with this symbolic expression,
         regardless of role.
         """
 
-        return ()
+        raise NotImplementedError  # pragma: no cover
 
-    def _attributes_repr(self):
-        # type: () -> str
+    def deep_eq(self, other: object) -> bool:
+        raise NotImplementedError  # pragma: no cover
+
+    def _attributes_repr(self) -> str:
         if not self.attributes:
             return "set()"
         else:
@@ -79,10 +82,6 @@ class SymbolicExpression:
             )
 
 
-if typing.TYPE_CHECKING:
-    AttributesCtorType = typing.Iterable[SymbolicExpression.Attribute]
-
-
 class SymAddrAddr(SymbolicExpression):
     """Represents a symbolic expression of the form
     "(Sym1 - Sym2) / Scale + Offset".
@@ -95,13 +94,12 @@ class SymAddrAddr(SymbolicExpression):
 
     def __init__(
         self,
-        scale,  # type: int
-        offset,  # type: int
-        symbol1,  # type: Symbol
-        symbol2,  # type: Symbol
-        attributes=set(),  # type: AttributesCtorType
+        scale: int,
+        offset: int,
+        symbol1: Symbol,
+        symbol2: Symbol,
+        attributes: AttributesCtorType = set(),
     ):
-        # type: (...) -> None
         """
         :param scale: Constant scale factor.
         :param offset: Constant offset.
@@ -111,26 +109,34 @@ class SymAddrAddr(SymbolicExpression):
             expression.
         """
         super().__init__(attributes)
-        self.scale = scale  # type: int
-        self.offset = offset  # type: int
-        self.symbol1 = symbol1  # type: Symbol
-        self.symbol2 = symbol2  # type: Symbol
+        self.scale = scale
+        self.offset = offset
+        self.symbol1 = symbol1
+        self.symbol2 = symbol2
 
     @classmethod
     def _from_protobuf(
         cls,
-        proto_symaddraddr,  # type: SymbolicExpression_pb2.SymAddrAddr
-        get_by_uuid,  # type: typing.Callable[[UUID], Node]
-    ):
-        # type: (...) -> SymAddrAddr
-        symbol1 = get_by_uuid(UUID(bytes=proto_symaddraddr.symbol1_uuid))
-        symbol2 = get_by_uuid(UUID(bytes=proto_symaddraddr.symbol2_uuid))
+        proto_symaddraddr: SymbolicExpression_pb2.SymAddrAddr,
+        get_by_uuid: typing.Callable[[UUID], typing.Optional[Node]],
+    ) -> "SymAddrAddr":
+        symbol1_uuid = UUID(bytes=proto_symaddraddr.symbol1_uuid)
+        symbol1 = get_by_uuid(symbol1_uuid)
+        if not isinstance(symbol1, Symbol):
+            raise DeserializationError(
+                "SymAddrAddr: UUID %s is not a Symbol" % symbol1_uuid
+            )
+        symbol2_uuid = UUID(bytes=proto_symaddraddr.symbol2_uuid)
+        symbol2 = get_by_uuid(symbol2_uuid)
+        if not isinstance(symbol2, Symbol):
+            raise DeserializationError(
+                "SymAddrAddr: UUID %s is not a Symbol" % symbol2_uuid
+            )
         return cls(
             proto_symaddraddr.scale, proto_symaddraddr.offset, symbol1, symbol2
         )
 
-    def _to_protobuf(self):
-        # type: () -> SymbolicExpression_pb2.SymAddrAddr
+    def _to_protobuf(self) -> SymbolicExpression_pb2.SymAddrAddr:
         proto_symaddraddr = SymbolicExpression_pb2.SymAddrAddr()
         proto_symaddraddr.scale = self.scale
         proto_symaddraddr.offset = self.offset
@@ -138,8 +144,7 @@ class SymAddrAddr(SymbolicExpression):
         proto_symaddraddr.symbol2_uuid = self.symbol2.uuid.bytes
         return proto_symaddraddr
 
-    def __eq__(self, other):
-        # type: (typing.Any) -> bool
+    def __eq__(self, other: object) -> bool:
         if not isinstance(other, SymAddrAddr):
             return False
         return (
@@ -150,14 +155,12 @@ class SymAddrAddr(SymbolicExpression):
             and self.attributes == other.attributes
         )
 
-    def __hash__(self):
-        # type: () -> int
+    def __hash__(self) -> int:
         return hash(
             (self.offset, self.scale, self.symbol1.uuid, self.symbol2.uuid)
         )
 
-    def __repr__(self):
-        # type: () -> str
+    def __repr__(self) -> str:
         return (
             "SymAddrAddr("
             "scale={scale!r}, "
@@ -166,10 +169,15 @@ class SymAddrAddr(SymbolicExpression):
             "symbol2={symbol2!r}, "
             "attributes={attributes_repr!s}, "
             ")"
-        ).format(attributes_repr=self._attributes_repr(), **self.__dict__)
+        ).format(
+            scale=self.scale,
+            offset=self.offset,
+            symbol1=self.symbol1,
+            symbol2=self.symbol2,
+            attributes_repr=self._attributes_repr(),
+        )
 
-    def deep_eq(self, other):
-        # type: (typing.Any) -> bool
+    def deep_eq(self, other: object) -> bool:
         # Do not move __eq__. See docstring for Node.deep_eq for more info.
         if not isinstance(other, SymAddrAddr):
             return False
@@ -182,7 +190,7 @@ class SymAddrAddr(SymbolicExpression):
         )
 
     @property
-    def symbols(self):
+    def symbols(self) -> typing.Iterable[Symbol]:
         yield self.symbol1
         yield self.symbol2
 
@@ -194,8 +202,12 @@ class SymAddrConst(SymbolicExpression):
     :ivar ~.symbol: Symbol representing an address.
     """
 
-    def __init__(self, offset, symbol, attributes=set()):
-        # type: (int, Symbol, AttributesCtorType) -> None
+    def __init__(
+        self,
+        offset: int,
+        symbol: Symbol,
+        attributes: AttributesCtorType = set(),
+    ):
         """
         :param offset: Constant offset.
         :param symbol: Symbol representing an address.
@@ -203,29 +215,31 @@ class SymAddrConst(SymbolicExpression):
             expression.
         """
         super().__init__(attributes)
-        self.offset = offset  # type: int
-        self.symbol = symbol  # type: Symbol
+        self.offset = offset
+        self.symbol = symbol
 
     @classmethod
     def _from_protobuf(
         cls,
-        proto_symaddrconst,  # type: SymbolicExpression_pb2.SymAddrConst
-        get_by_uuid,  # type: typing.Callable[[UUID], Node]
-    ):
-        # type: (...) -> SymAddrConst
-        symbol = get_by_uuid(UUID(bytes=proto_symaddrconst.symbol_uuid))
+        proto_symaddrconst: SymbolicExpression_pb2.SymAddrConst,
+        get_by_uuid: typing.Callable[[UUID], typing.Optional[Node]],
+    ) -> "SymAddrConst":
+        symbol_uuid = UUID(bytes=proto_symaddrconst.symbol_uuid)
+        symbol = get_by_uuid(symbol_uuid)
+        if not isinstance(symbol, Symbol):
+            raise DeserializationError(
+                "SymAddrConst: UUID %s is not a Symbol" % symbol_uuid
+            )
         return cls(proto_symaddrconst.offset, symbol)
 
-    def _to_protobuf(self):
-        # type: () -> SymbolicExpression_pb2.SymAddrConst
+    def _to_protobuf(self) -> SymbolicExpression_pb2.SymAddrConst:
         proto_symaddrconst = SymbolicExpression_pb2.SymAddrConst()
         proto_symaddrconst.offset = self.offset
         if self.symbol is not None:
             proto_symaddrconst.symbol_uuid = self.symbol.uuid.bytes
         return proto_symaddrconst
 
-    def __eq__(self, other):
-        # type: (typing.Any) -> bool
+    def __eq__(self, other: object) -> bool:
         if not isinstance(other, SymAddrConst):
             return False
         return (
@@ -234,22 +248,23 @@ class SymAddrConst(SymbolicExpression):
             and self.attributes == other.attributes
         )
 
-    def __hash__(self):
-        # type: () -> int
+    def __hash__(self) -> int:
         return hash((self.offset, self.symbol.uuid))
 
-    def __repr__(self):
-        # type: () -> str
+    def __repr__(self) -> str:
         return (
             "SymAddrConst("
             "offset={offset!r}, "
             "symbol={symbol!r}, "
             "attributes={attributes_repr!s}, "
             ")"
-        ).format(attributes_repr=self._attributes_repr(), **self.__dict__)
+        ).format(
+            offset=self.offset,
+            symbol=self.symbol,
+            attributes_repr=self._attributes_repr(),
+        )
 
-    def deep_eq(self, other):
-        # type: (typing.Any) -> bool
+    def deep_eq(self, other: object) -> bool:
         # Do not move __eq__. See docstring for Node.deep_eq for more info.
         if not isinstance(other, SymAddrConst):
             return False
@@ -260,5 +275,5 @@ class SymAddrConst(SymbolicExpression):
         )
 
     @property
-    def symbols(self):
+    def symbols(self) -> typing.Iterable[Symbol]:
         yield self.symbol
